#include "AxeParser.h"

#include "llvm/DerivedTypes.h"
#include "llvm/Support/Casting.h"

int main(int argc, char **argv) {
	InputParser input(argc, argv);

	switch (input.getType()) {
		case InputParser::NONE:
			fputs("usage: ", stdout);
			fputs(argv[0], stdout);
			fputs(" <input-file> <output-file>\n", stdout);
			return 0;
		case InputParser::INVALID:
			fputs("Invalid input file!\n", stdout);
			return 1;
		case InputParser::TEXT:
			fputs("Text files are not supported yet.\n", stdout);
			return 0;
		case InputParser::PROGRAM:
			break;
	}

	AxeParser parser(input);

	if (!parser.check()) {
		fputs("Invalid axe header!\n", stdout);
		return 2;
	}

	printf("Compiling %s to %s...\n", input.getName(), parser.getName());

	if (!parser.parse()) {
		if (parser.getError() != NULL)
			printf("Error: %s\nCurrent: 0x%02hx\n", parser.getError(),
					(unsigned short)input.currentToken());
		else
			fputs("Unknown Error!\n", stdout);
		return 3;
	}

	return 0;
}

AxeParser::AxeParser(InputParser &inputParser) :
		error(NULL), input(inputParser), module(NULL), builder(NULL),
		indent(0) { }

AxeParser::~AxeParser() {
	delete builder;
	delete module;
}

bool AxeParser::check() {
	int i = 0, t = input.nextToken();
	if (t != 0x3a || input.eof())
		return false;
	t = input.nextToken();
	if (t < 0x40 || t > 0x5b || input.eof())
		return false;
	do {
		name[i] = t;
		++i;
		t = input.nextToken();
	} while (t >= 0x40 && t <= 0x5b && i < 8 && !input.eof());
	name[i] = 0;
	while (t != 0x3f && !input.eof())
		t = input.nextToken();

	input.setOutput(name);
	module = new Module(name, getGlobalContext());
	return true;
}

#if 0
bool AxeParser::parseComment() {
	error = ERR_UNSUPPORTED;
	return false;
}

bool AxeParser::parseNewLine() {
	input.nextToken(); // consume
	return true;
}

bool AxeParser::parseBinaryConstant() {
	error = ERR_UNSUPPORTED;
	return false;
}

bool AxeParser::parseHexadecimalConstant() {
	error = ERR_UNSUPPORTED;
	return false;
}

bool AxeParser::parseDecimalConstant() {
	unsigned short value = 0;
	do {
		unsigned short newValue = value * 10 + input.currentToken() - 0x30;
		if (newValue < value) {
			error = ERR_BIG_CONSTANT;
			return false;
		}
		value = newValue;
	} while (input.nextToken() >= 0x30 && input.currentToken() <= 0x39);
	if (input.currentToken() == 0x3a) {
		error = ERR_UNSUPPORTED;
		return false;
	}
	ans = ConstantInt::get(getGlobalContext(), APInt(16, value));
	return true;
}

bool AxeParser::parseStringConstant() {
	while (input.nextToken() != 0x2a && input.currentToken() != 0x3f &&
			input.currentToken() != 0x04) {
	}
}

bool AxeParser::parseNext() {
	switch (input.currentToken()) {
		case 0x30: // 0
		case 0x31: // 1
		case 0x32: // 2
		case 0x33: // 3
		case 0x34: // 4
		case 0x35: // 5
		case 0x36: // 6
		case 0x37: // 7
		case 0x38: // 8
		case 0x39: // 9
			return parseDecimalConstant();
		case 0x3a: // .
			return parseComment();
		case 0x3b: // small E
			return parseHexadecimalConstant();
		case 0x3f: // return
			return parseNewLine();
		case 0xac: // pi
			return parseBinaryConstant();
		default:
			error = ERR_UNKNOWN;
			return false;
	}
}

Value *AxeParser::parseConstantExpression() {
	switch (input.currentToken()) {
		case 0x06: // [
		case 0x30: // 0
		case 0x31: // 1
		case 0x32: // 2
		case 0x33: // 3
		case 0x34: // 4
		case 0x35: // 5
		case 0x36: // 6
		case 0x37: // 7
		case 0x38: // 8
		case 0x39: // 9
		case 0x3b: // small E
		case 0xac: // pi
	}
}

// Operation := ( [+-*/^=≠<>≤≥ or  and  xor ] | [ ) Expression | [→] Identifier ( [ʳ] )
Value *AxeParser::parseOperation(Value *op) {
	if (input.currentToken() == 0x04)
		return parseError("assign");
	
}

// Constant := 
Value *AxeParser::parseConstant() {
	
}

// ShortList := Expression

// List := [(] ShortList ( [,] ShortList ( [,] ShortList ) ) [)]

// Identifier := [A-θ] ( [0-9A-θa-z] ( [0-9A-θa-z] ( [0-9A-θa-z] ( [0-9A-θa-z] ) ) ) )
Value *AxeParser::parseIdentifier() {
	
}

// Value := [-] Value | [(] Expression ( [)] ) | Identifier ( { [ ] } [(] List ( [)] ) ) | Constant
Value *AxeParser::parseValue() {
	if (input.currentToken() == 0xb0) {
		nextToken(); // consume
		Value *op = parseValue();
	}
	if (input.currentToken() >= 0x41 && input.currentToken() <= 0x5b)
		return parseIdentifier();
	else
		return parseConstant();
}

// Expression := Value { Operation }
Value *AxeParser::parseExpression() {
	Value *op = parseValue();
	parseOperation(op)
}
#endif

bool AxeParser::parse() {
	if (module == NULL)
		return false;

	init = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "init",
			module);
	getkey = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), false), Function::ExternalLinkage, "getkey",
			module);
	rand = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), false), Function::ExternalLinkage, "rnd",
			module);
	pause = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), Type::getInt16Ty(getGlobalContext()), false),
			Function::ExternalLinkage, "pause", module);
	normal = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "normal",
			module);
	full = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "full",
			module);
	inton = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "inton",
			module);
	intoff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "intoff",
			module);
	diagon = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "diagon",
			module);
	diagoff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "diagoff",
			module);
	clrfront = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "clrfront",
			module);
	clrback = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "clrback",
			module);
	Type *pxlArgs[] = {
			Type::getInt8Ty(getGlobalContext()),
			Type::getInt8Ty(getGlobalContext())
	};
	pxlon = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "pxlon", module);
	pxloff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "pxloff", module);
	pxlchange = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "pxlchange", module);
	pxltest = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "pxltest", module);
	backpxlon = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "backpxlon", module);
	backpxloff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "backpxloff", module);
	backpxlchange = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "backpxlchange", module);
	backpxltest = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), pxlArgs, false),
			Function::ExternalLinkage, "backpxltest", module);
	Type *spriteArgs[] = {
			Type::getInt8Ty(getGlobalContext()),
			Type::getInt8Ty(getGlobalContext()),
			Type::getInt8PtrTy(getGlobalContext())
	};
	spriteon = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "spriteon", module);
	spriteoff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "spriteoff", module);
	spritechange = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "spritechange", module);
	spriteget = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "spriteget", module);
	backspriteon = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "backspriteon", module);
	backspriteoff = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "backspriteoff", module);
	backspritechange = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "backspritechange", module);
	backspriteget = Function::Create(FunctionType::get(Type::getInt16Ty(
			getGlobalContext()), spriteArgs, false),
			Function::ExternalLinkage, "backspriteget", module);
	dispgraph = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "dispgraph",
			module);
	clrhome = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), false), Function::ExternalLinkage, "clrhome",
			module);
	dispstr = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), Type::getInt8PtrTy(getGlobalContext()), false),
			Function::ExternalLinkage, "dispstr", module);
	Type *outputArgs[] = {
			Type::getInt8Ty(getGlobalContext()),
			Type::getInt8Ty(getGlobalContext()),
			Type::getInt8PtrTy(getGlobalContext())
	};
	output = Function::Create(FunctionType::get(Type::getVoidTy(
			getGlobalContext()), outputArgs, false), Function::ExternalLinkage,
			"output", module);
	builder = new IRBuilder<>(BasicBlock::Create(getGlobalContext(), "entry",
			Function::Create(FunctionType::get(
					Type::getInt16Ty(getGlobalContext()), false),
					Function::ExternalLinkage, "start", module)));

	{
		char identifier[2] = { 0, 0 };
		for (*identifier = 0x41; *identifier <= 0x5b; ++*identifier)
			variables[identifier] = new GlobalVariable(*module,
					builder->getInt16Ty(), false,
					GlobalVariable::PrivateLinkage, UndefValue::get(
					builder->getInt16Ty()), identifier);
	}

	//input.nextToken();
	if (!parseProgram())
		return false;

	module->dump();

	return true;
}

#ifdef DEBUG
#define Enter(str) printf("%s%s\n", std::string(indent++, ' ').c_str(), str)
#define PrintF(format, value) printf("%s"format"\n", std::string(indent, ' ') \
		.c_str(), value)
#define Exit() --indent
#else
#define Enter(str)
#define PrintF(format, value)
#define Exit()
#endif

// Program := Block
bool AxeParser::parseProgram() {
	builder->CreateCall(init);
	Enter("Program");
		Value *current = UndefValue::get(builder->getInt16Ty());
		Enter("Block");
			current = parseBlock(current);
		Exit();
	Exit();
	if (!current)
		return false;
	builder->CreateRet(UndefValue::get(builder->getInt16Ty()));
	return input.eofCurrent();
}

// Block := Statement { 1 [\n] } ( Block )
Value *AxeParser::parseBlock(Value *current) {
	if (!current || input.currentToken() == 0xd0 ||
			input.currentToken() == 0xd4)
		return current;
	Enter("Statement");
		current = parseStatement(current);
	Exit();
	if (!current || input.currentToken() == 0xd0 ||
			input.currentToken() == 0xd4)
		return current;
	if (input.eofCurrent())
		return current;
	if (!input.consumeToken(0x3e) && !input.consumeToken(0x3f))
		return parseError(ERR_UNKNOWN);
	return parseBlock(current);
}

// Statement := If | Loop | Return | Expression
// If := ( [!] ) [If ] Expression
// Loop := [Lbl ] Label | [Goto ] Label | [Repeat ] Expression | [While ] Expression
// Return := [Return] ( If | Expression )
Value *AxeParser::parseStatement(Value *current) {
	Enter("Expression");
		current = parseExpression(current);
	Exit();
	switch (input.currentToken()) {
		//case 0x3e: // :
			//input.consumeToken();
			//break;
		//case 0x3f: // \n
			//return current;
		case 0x66: // Normal
			input.consumeToken();
			Enter("Normal");
				builder->CreateCall(normal);
			Exit();
			return current;
		case 0x75: // Full
			input.consumeToken();
			Enter("Full");
				builder->CreateCall(full);
			Exit();
			return builder->getInt16(-1);
		case 0x85: // ClrDraw
			input.consumeToken();
			Enter("ClrDraw");
				if (input.consumeToken(0x0a))
					builder->CreateCall(clrback);
				else
					builder->CreateCall(clrfront);
			Exit();
			return current;
		case 0x96: // FnOn
			input.consumeToken();
			Enter("Interrupts On");
				builder->CreateCall(inton);
			Exit();
			return current;
		case 0x97: // FnOff
			input.consumeToken();
			Enter("Interrupts Off");
				builder->CreateCall(intoff);
			Exit();
			return current;
		case 0xbb66: // DiagnosticOn
			input.consumeToken();
			Enter("DiagnosticOn");
				builder->CreateCall(diagon);
			Exit();
			return current;
		case 0xbb67: // DiagnosticOff
			input.consumeToken();
			Enter("DiagnosticOff");
				builder->CreateCall(diagoff);
			Exit();
			return current;
		case 0x2d: // !
		case 0xce: // If
			Enter("If");
				current = parseIfStatement(current);
			Exit();
			return current;
		case 0xd1: // While 
		case 0xd2: // Repeat 
			Enter("Loop");
				current = parseLoopStatement(current);
			Exit();
			return current;
		case 0xd3: // For(
			input.consumeToken();
			Enter("For");
				current = parseForStatement(current);
			Exit();
			return current;
		case 0xd5: // Return
			input.consumeToken();
			Enter("Return");
				current = parseReturnStatement(current);
			Exit();
			return current;
		case 0xd6: // Lbl 
			input.consumeToken();
			Enter("Lbl");
				current = parseLblStatement(current);
			Exit();
			return current;
		case 0xd7: // Goto 
			input.consumeToken();
			Enter("Goto");
				current = parseGotoStatement(current);
			Exit();
			return current;
		case 0xd8: // Pause 
			input.consumeToken();
			Enter("Pause");
				Enter("Value");
					current = parseValue(current);
				Exit();
				builder->CreateCall(pause, current);
			Exit();
			return builder->getInt16(0);
		case 0xde: // Disp 
			input.consumeToken();
			Enter("Disp");
				Enter("Value");
					current = parseValue(current);
				Exit();
				current = builder->CreatePointerCast(current,
						builder->getInt8PtrTy());
				builder->CreateCall(dispstr, current);
			Exit();
			return UndefValue::get(builder->getInt16Ty());
		case 0xdf: // DispGraph
			input.consumeToken();
			Enter("DispGraph");
				builder->CreateCall(dispgraph);
			Exit();
			input.consumeToken(0x0a); // TODO: Implement
			return UndefValue::get(builder->getInt16Ty());
		case 0xe0: // Output(
			input.consumeToken();
			Enter("Output(");
				current = parseOutput(current);
			Exit();
			return current;
		case 0xe1: // ClrHome
			input.consumeToken();
			Enter("Clear Home");
				builder->CreateCall(clrhome);
			Exit();
			return UndefValue::get(builder->getInt16Ty());
		default:
			return current;
	}
	if (!current || input.eofCurrent())
		return current;
	return parseStatement(current);
}

// IfStatement := ( [!] ) [If ] Value Block
Value *AxeParser::parseIfStatement(Value *current) {
	bool type = input.consumeToken(0x2d);
	if (!input.consumeToken(0xce))
		return parseError(ERR_NO_IF);
	Enter("Value");
		current = parseValue(current);
	Exit();
	if (!current)
		return current;
	BasicBlock *thenBlock = BasicBlock::Create(getGlobalContext(), "then",
			builder->GetInsertBlock()->getParent());
	BasicBlock *elseBlock = BasicBlock::Create(getGlobalContext(), "else",
			builder->GetInsertBlock()->getParent());
	BasicBlock *endBlock = BasicBlock::Create(getGlobalContext(), "endif",
			builder->GetInsertBlock()->getParent());
	Value *condition;
	if (type)
		condition = builder->CreateICmpEQ(current, builder->getInt16(0));
	else
		condition = builder->CreateICmpNE(current, builder->getInt16(0));
	builder->CreateCondBr(condition, thenBlock, elseBlock);
	builder->SetInsertPoint(endBlock);
	PHINode *phi = builder->CreatePHI(builder->getInt16Ty(), 2);
	builder->SetInsertPoint(thenBlock);
	Enter("Block");
		Value *trueCurrent = parseBlock(current);
	Exit();
	if (!trueCurrent)
		return trueCurrent;
	builder->CreateBr(endBlock);
	phi->addIncoming(trueCurrent, builder->GetInsertBlock());
	builder->SetInsertPoint(elseBlock);
	if (input.consumeToken(0xd0)) {
		Enter("Block");
			current = parseBlock(current);
		Exit();
		if (!current)
			return current;
	}
	if (!input.consumeToken(0xd4))
		return parseError(ERR_NO_END);
	builder->CreateBr(endBlock);
	phi->addIncoming(current, builder->GetInsertBlock());
	builder->SetInsertPoint(endBlock);
	return phi;
}

// LoopStatement := [While ] | [Repeat ] Value Block
Value *AxeParser::parseLoopStatement(Value *current) {
	bool type;
	if (input.consumeToken(0xd1))
		type = true; // While 
	else if (input.consumeToken(0xd2))
		type = false; // Repeat 
	else
		return parseError(ERR_NO_LOOP);
	BasicBlock *condBlock = BasicBlock::Create(getGlobalContext(), "loopcond",
			builder->GetInsertBlock()->getParent());
	BasicBlock *bodyBlock = BasicBlock::Create(getGlobalContext(), "loopbody",
			builder->GetInsertBlock()->getParent());
	BasicBlock *endBlock = BasicBlock::Create(getGlobalContext(), "loopend",
			builder->GetInsertBlock()->getParent());
	builder->CreateBr(condBlock);
	BasicBlock *incomingBlock = builder->GetInsertBlock();
	builder->SetInsertPoint(condBlock);
	PHINode *phi = builder->CreatePHI(builder->getInt16Ty(), 2);
	phi->addIncoming(current, incomingBlock);
	Enter("Value");
		Value *condition = parseValue(phi);
	Exit();
	if (!condition)
		return condition;
	if (!isa<Constant>(condition))
		current = condition;
	else
		current = phi;
	condition = builder->CreateICmpNE(condition, builder->getInt16(0));
	builder->CreateCondBr(condition, type ? bodyBlock : endBlock,
			type ? endBlock : bodyBlock);
	builder->SetInsertPoint(bodyBlock);
	Enter("Block");
		current = parseBlock(current);
	Exit();
	if (!current)
		return current;
	if (!input.consumeToken(0xd4))
		return parseError(ERR_NO_END);
	builder->CreateBr(condBlock);
	phi->addIncoming(current, builder->GetInsertBlock());
	builder->SetInsertPoint(endBlock);
	return current;
}

// ForStatement := [For(] Args Block
Value *AxeParser::parseForStatement(Value *current) {
	BasicBlock *condBlock = BasicBlock::Create(getGlobalContext(), "forcond",
			builder->GetInsertBlock()->getParent());
	BasicBlock *bodyBlock = BasicBlock::Create(getGlobalContext(), "forbody",
			builder->GetInsertBlock()->getParent());
	BasicBlock *updateBlock = BasicBlock::Create(getGlobalContext(), "forup",
			builder->GetInsertBlock()->getParent());
	BasicBlock *endBlock = BasicBlock::Create(getGlobalContext(), "forend",
			builder->GetInsertBlock()->getParent());
	bool type;
	if (input.currentToken() >= 0x41 && input.currentToken() <= 0x5b) {
		Enter("Load");
			Enter("Identifier");
				Value *variable = parseIdentifier();
			Exit();
		Exit();
		if (!variable)
			return variable;
		if ((type = input.consumeToken(0x2b))) {
			Enter("Expression");
				current = parseExpression(current);
			Exit();
			if (!current)
				return current;
			if (!input.consumeToken(0x2b))
				return parseError(ERR_WRONG_ARGS);
			builder->CreateStore(current, variable);
			builder->CreateBr(condBlock);
			builder->SetInsertPoint(updateBlock);
			builder->CreateStore(builder->CreateAdd(
					builder->CreateLoad(variable), builder->getInt16(1)),
					variable);
			builder->CreateBr(condBlock);
			builder->SetInsertPoint(condBlock);
			Value *next = builder->CreateLoad(variable);
			builder->CreateStore(next, variable);
			current = UndefValue::get(builder->getInt16Ty());
			Enter("Expression");
				Value *done = parseExpression(current);
			Exit();
			if (!done)
				return done;
			Value *condition = builder->CreateICmpULE(next, done);
			builder->CreateCondBr(condition, bodyBlock, endBlock);
		} else current = builder->CreateLoad(variable);
	} else type = false;
	PHINode *currentPHI;
	if (!type) {
		Enter("Expression");
			Value *times = parseExpression(current);
		Exit();
		if (!times)
			return times;
		if (!isa<Constant>(times))
			current = times;
		builder->CreateBr(condBlock);
		BasicBlock *incomingBlock = builder->GetInsertBlock();
		builder->SetInsertPoint(condBlock);
		PHINode *phi = builder->CreatePHI(builder->getInt16Ty(), 2);
		currentPHI = builder->CreatePHI(builder->getInt16Ty(), 2);
		Value *newTimes = builder->CreateSub(phi, builder->getInt16(1));
		Value *condition = builder->CreateICmpEQ(newTimes,
				builder->getInt16(0));
		builder->CreateCondBr(condition, endBlock, bodyBlock);
		phi->addIncoming(times, incomingBlock);
		phi->addIncoming(newTimes, bodyBlock);
		currentPHI->addIncoming(current, incomingBlock);
		current = currentPHI;
	}
	input.consumeToken(0x11);
	builder->SetInsertPoint(bodyBlock);
	Enter("Block");
		Value *bodyCurrent = parseBlock(current);
	Exit();
	if (!bodyCurrent)
		return bodyCurrent;
	if (!input.consumeToken(0xd4))
		return parseError(ERR_NO_END);
	builder->CreateBr(updateBlock);
	builder->SetInsertPoint(endBlock);
	if (!type)
		currentPHI->addIncoming(bodyCurrent, bodyBlock);
	return current;
}

// LblStatement := 
Value *AxeParser::parseLblStatement(Value *current) {
	Enter("Label");
		Function *label = parseLabel();
	Exit();
	builder->CreateRet(builder->CreateCall(label, current));
	builder->SetInsertPoint(BasicBlock::Create(getGlobalContext(), "entry",
			label));
	return label->arg_begin();
}

// GotoStatement := 
Value *AxeParser::parseGotoStatement(Value *current) {
	Enter("Label");
		Function *label = parseLabel();
	Exit();
	builder->CreateRet(builder->CreateCall(label, current));
	builder->SetInsertPoint(BasicBlock::Create(getGlobalContext(),
			"nogoto", builder->GetInsertBlock()->getParent()));
	return UndefValue::get(builder->getInt16Ty());
}

// ReturnStatement := ( ( [!] ) [If ] ) Value
Value *AxeParser::parseReturnStatement(Value *current) {
	bool type = input.consumeToken(0x2d);
	bool cond = input.consumeToken(0xce);
	Enter("Value");
		current = parseValue(current);
	Exit();
	if (!current)
		return current;
	BasicBlock *endBlock = BasicBlock::Create(getGlobalContext(), "noret",
			builder->GetInsertBlock()->getParent());
	if (cond) {
		BasicBlock *retBlock = BasicBlock::Create(getGlobalContext(), "ret",
				builder->GetInsertBlock()->getParent());
		Value *condition;
		if (type)
			condition = builder->CreateICmpEQ(current, builder->getInt16(0));
		else
			condition = builder->CreateICmpNE(current, builder->getInt16(0));
		builder->CreateCondBr(condition, retBlock, endBlock);
		builder->SetInsertPoint(retBlock);
	}
	builder->CreateRet(current);
	builder->SetInsertPoint(endBlock);
	return current;
}

// Expression := Value { 1 [:] } ( Expression )
Value *AxeParser::parseExpression(Value *current) {
	Enter("Value");
		current = parseValue(current);
	Exit();
	if (!current || !input.consumeToken(0x3e))
		return current;
	return parseExpression(current);
}

// Value := { ( [->+-*/^] ) Operand }
Value *AxeParser::parseValue(Value *current) {
	bool lookahead = true;
	while (lookahead && current) switch (input.currentToken()) {
		case 0x04: // ->
			input.consumeToken();
			Enter("Store");
				current = parseStore(current);
			Exit();
			break;
		case 0x3c: //  or 
			input.consumeToken();
			Enter("Or");
				current = parseOr(current);
			Exit();
			break;
		case 0x3d: //  xor 
			input.consumeToken();
			Enter("Xor");
				current = parseXor(current);
			Exit();
			break;
		case 0x40: //  and 
			input.consumeToken();
			Enter("And");
				current = parseAnd(current);
			Exit();
			break;
		case 0x6a: // =
			input.consumeToken();
			Enter("Equal");
				current = parseEqual(current);
			Exit();
			break;
		case 0x6b: // <
			input.consumeToken();
			Enter("Less Than");
				current = parseLessThan(current);
			Exit();
			break;
		case 0x6c: // >
			input.consumeToken();
			Enter("Greater Than");
				current = parseGreaterThan(current);
			Exit();
			break;
		case 0x6d: // ≤
			input.consumeToken();
			Enter("Less Equal");
				current = parseLessEqual(current);
			Exit();
			break;
		case 0x6e: // ≥
			input.consumeToken();
			Enter("Greater Equal");
				current = parseGreaterEqual(current);
			Exit();
			break;
		case 0x6f: // ≠
			input.consumeToken();
			Enter("Not Equal");
				current = parseNotEqual(current);
			Exit();
			break;
		case 0x70: // +
			input.consumeToken();
			Enter("Add");
				current = parseAdd(current);
			Exit();
			break;
		case 0x71: // -
			input.consumeToken();
			Enter("Subtract");
				current = parseSubtract(current);
			Exit();
			break;
		case 0x82: // *
			input.consumeToken();
			Enter("Multiply");
				current = parseMultiply(current);
			Exit();
			break;
		case 0x83: // /
			input.consumeToken();
			Enter("Divide");
				current = parseDivide(current);
			Exit();
			break;
		case 0xb0: // -
			input.consumeToken();
			Enter("Negate");
				current = parseNegate(current);
			Exit();
			break;
		case 0xf0: // ^
			input.consumeToken();
			Enter("Modulus");
				current = parseModulus(current);
			Exit();
			break;
		default:
			Enter("Operand");
				Value *next = parseOperand(current, &lookahead);
			Exit();
			if (lookahead)
				current = next;
			break;
	}
	return current;
}

// Operand := [(] Expression ( [)] ) | ...
Value *AxeParser::parseOperand(Value *current, bool *lookahead) {
	if (lookahead) *lookahead = true;
	switch (input.currentToken()) {
		case 0x06:
		case 0x2a:
			Enter("Constant Data");
				current = parseConstantData();
			Exit();
			return current;
		case 0x10: // (
			input.consumeToken();
			Enter("Expression");
				current = parseExpression(current);
			Exit();
			input.consumeToken(0x11);
			return current;
		case 0x13: // pxl-Test(
			input.consumeToken();
			Enter("Pixel Test");
				current = parsePxlTest(current);
			Exit();
			return current;
		case 0x30: // 0
		case 0x31: // 1
		case 0x32: // 2
		case 0x33: // 3
		case 0x34: // 4
		case 0x35: // 5
		case 0x36: // 6
		case 0x37: // 7
		case 0x38: // 8
		case 0x39: // 9
			Enter("Decimal Constant");
				current = parseDecimalConstant();
			Exit();
			return current;
		case 0x3b: // lower E
			Enter("Hexadecimal Constant");
				current = parseHexadecimalConstant();
			Exit();
			return current;
		case 0x41: // A
		case 0x42: // B
		case 0x43: // C
		case 0x44: // D
		case 0x45: // E
		case 0x46: // F
		case 0x47: // G
		case 0x48: // H
		case 0x49: // I
		case 0x4a: // J
		case 0x4b: // K
		case 0x4c: // L
		case 0x4d: // M
		case 0x4e: // N
		case 0x4f: // O
		case 0x50: // P
		case 0x51: // Q
		case 0x52: // R
		case 0x53: // S
		case 0x54: // T
		case 0x55: // U
		case 0x56: // V
		case 0x57: // W
		case 0x58: // X
		case 0x59: // Y
		case 0x5a: // Z
		case 0x5b: // θ
		case 0x6000: // Pic1
		case 0x6001: // Pic2
		case 0x6002: // Pic3
		case 0x6003: // Pic4
		case 0x6004: // Pic5
		case 0x6005: // Pic6
		case 0x6006: // Pic7
		case 0x6007: // Pic8
		case 0x6008: // Pic9
		case 0x6009: // Pic0
		case 0x6100: // GDB1
		case 0x6101: // GDB2
		case 0x6102: // GDB3
		case 0x6103: // GDB4
		case 0x6104: // GDB5
		case 0x6105: // GDB6
		case 0x6106: // GDB7
		case 0x6107: // GDB8
		case 0x6108: // GDB9
		case 0x6109: // GDB0
		case 0xaa00: // Str1
		case 0xaa01: // Str2
		case 0xaa04: // Str5
		case 0xaa02: // Str3
		case 0xaa03: // Str4
		case 0xaa05: // Str6
		case 0xaa06: // Str7
		case 0xaa07: // Str8
		case 0xaa08: // Str9
		case 0xaa09: // Str0
			Enter("Load");
				current = parseLoad();
			Exit();
			return current;
		case 0x9e: // Pt-On(
			input.consumeToken();
			Enter("Sprite On");
				current = parsePtOn(current);
			Exit();
			return current;
		case 0x9f: // Pt-Off(
			input.consumeToken();
			Enter("Sprite Off");
				current = parsePtOff(current);
			Exit();
			return current;
		case 0xa0: // Pt-Change(
			input.consumeToken();
			Enter("Sprite Change");
				current = parsePtChange(current);
			Exit();
			return current;
		case 0xa1: // Pxl-On(
			input.consumeToken();
			Enter("Pixel On");
				current = parsePxlOn(current);
			Exit();
			return current;
		case 0xa2: // Pxl-Off(
			input.consumeToken();
			Enter("Pixel Off");
				current = parsePxlOff(current);
			Exit();
			return current;
		case 0xa3: // Pxl-Change(
			input.consumeToken();
			Enter("Pixel Change");
				current = parsePxlChange(current);
			Exit();
			return current;
		case 0xab: // rand
			input.consumeToken();
			Enter("Rand");
				current = builder->CreateCall(rand);
			Exit();
			return current;
		case 0xad: // getKey
			input.consumeToken();
			Enter("GetKey");
				current = parseGetKey(current);
			Exit();
			return current;
		case 0xbb0c: // sub(
			input.consumeToken();
			Enter("Sub");
				current = parseSub(current);
			Exit();
			return current;
		case 0xed: // pt-Get(
			input.consumeToken();
			Enter("Sprite Get");
				current = parsePtGet(current);
			Exit();
			return current;
		default:
			if (!lookahead)
				return parseError(ERR_UNKNOWN);
			*lookahead = false;
			return 0;
	}
}

// ConstantData := 
Constant *AxeParser::parseConstantData() {
	bool terminator;
	std::vector<Constant *> constants;
	while (true) switch (input.currentToken()) {
		case 0x06: // [
			while ((input.nextToken() >= 0x30 &&
					input.currentToken() <= 0x39) ||
					(input.currentToken() >= 0x41 &&
					input.currentToken() <= 0x46)) {
				Constant *constant = parseHexadecimalByte();
				if (!constant)
					return constant;
				constants.push_back(constant);
			}
			input.consumeToken(0x07);
			terminator = false;
			break;
		case 0x2a: // "
			input.consumeToken();
			while (!input.consumeToken(0x2a) && input.currentToken() != 0x04 &&
					input.currentToken() != 0x3f)
				constants.push_back(builder->getInt8(
						*TokenSet::TIOS[input.consumeToken()]));
			terminator = true;
			break;
		default:
			if (constants.empty())
				return parseError(ERR_NO_CONSTANT);
			if (terminator)
				constants.push_back(builder->getInt8(0));
			return new GlobalVariable(*module, ArrayType::get(builder->getInt8Ty(),
					constants.size()), false, GlobalVariable::PrivateLinkage,
					ConstantArray::get(ArrayType::get(builder->getInt8Ty(),
					constants.size()), constants), "global");
	}
}

// Sub := [sub(] Label ( [,] Args | [)] )
Value *AxeParser::parseSub(Value *current) {
	Enter("Label");
		Value *label = parseLabel();
	Exit();
	input.consumeToken(0x11);
	return builder->CreateCall(label, current);
}

// Output := [Output(] Args
Value *AxeParser::parseOutput(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 3)
		return parseError(ERR_UNSUPPORTED);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(output, args);
	current = args.back();
	if (!current)
		return current;
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pxl-On(] Args
Value *AxeParser::parsePxlOn(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 2)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	//args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backpxlon : pxlon, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pxl-Off(] Args
Value *AxeParser::parsePxlOff(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 2)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	//args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backpxloff : pxloff, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pxl-Change(] Args
Value *AxeParser::parsePxlChange(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 2)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	//args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backpxlchange : pxlchange, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [pxl-Test(] Args
Value *AxeParser::parsePxlTest(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 2)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	//args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	return builder->CreateCall(input.consumeToken(0x0a) ? backpxltest : pxltest, args);
}

// Output := [Pt-On(] Args
Value *AxeParser::parsePtOn(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 3)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backspriteon : spriteon, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pt-Off(] Args
Value *AxeParser::parsePtOff(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 3)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backspriteoff : spriteoff, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pt-Change(] Args
Value *AxeParser::parsePtChange(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 3)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	builder->CreateCall(input.consumeToken(0x0a) ? backspritechange : spritechange, args);
	return UndefValue::get(builder->getInt16Ty());
}

// Output := [Pt-Get(] Args
Value *AxeParser::parsePtGet(Value *current) {
	// TODO: implement
	Enter("Args");
		std::vector<Value *> args = parseArgs(current);
	Exit();
	if (args.size() != 3)
		return parseError(ERR_WRONG_ARGS);
	for (int i = 0; i < 2; ++i)
		args[i] = builder->CreateIntCast(args[i], builder->getInt8Ty(), false);
	args[2] = builder->CreatePointerCast(args[2], builder->getInt8PtrTy());
	return builder->CreateCall(input.consumeToken(0x0a) ? backspriteget : spriteget, args);
}

// GetKey := [getKey]
Value *AxeParser::parseGetKey(Value *current) {
	// TODO: implement
	return builder->CreateCall(getkey);
}

// Args := { Expression [,] } Expression ( [)] )
std::vector<Value *> AxeParser::parseArgs(Value *current) {
	std::vector<Value *> args;
	do {
		Enter("Expression");
			current = parseExpression(current);
		Exit();
		args.push_back(current);
		if (!current)
			return args;
	} while (input.consumeToken(0x2b));
	input.consumeToken(0x11);
	return args;
}

Value *AxeParser::parseEqual(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpEQ(current, operand),
			builder->getInt16Ty(), false);
}
Value *AxeParser::parseLessThan(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpULT(current, operand),
			builder->getInt16Ty(), false);
}
Value *AxeParser::parseGreaterThan(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpUGT(current, operand),
			builder->getInt16Ty(), false);
}
Value *AxeParser::parseLessEqual(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpULE(current, operand),
			builder->getInt16Ty(), false);
}
Value *AxeParser::parseGreaterEqual(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpUGE(current, operand),
			builder->getInt16Ty(), false);
}
Value *AxeParser::parseNotEqual(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateIntCast(builder->CreateICmpNE(current, operand),
			builder->getInt16Ty(), false);
}

Value *AxeParser::parseAdd(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateAdd(current, operand);
}

Value *AxeParser::parseSubtract(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateSub(current, operand);
}

Value *AxeParser::parseMultiply(Value *current) {
	int type;
	if (input.consumeToken(0x82))
		type = 2;
	else if (input.consumeToken(0xf0))
		type = 1;
	else
		type = 0;
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	if (!type)
		return builder->CreateMul(current, operand);
	else if (type & 1) {
		current = builder->CreateIntCast(current, builder->getInt32Ty(), false);
		operand = builder->CreateIntCast(operand, builder->getInt32Ty(), false);
		current = builder->CreateMul(current, operand);
		current = builder->CreateLShr(current, builder->getInt32(16));
		return builder->CreateIntCast(current, builder->getInt16Ty(), false);
	} else {
		current = builder->CreateIntCast(current, builder->getInt32Ty(), true);
		operand = builder->CreateIntCast(operand, builder->getInt32Ty(), true);
		current = builder->CreateMul(current, operand);
		current = builder->CreateLShr(current, builder->getInt32(8));
		return builder->CreateIntCast(current, builder->getInt16Ty(), false);
	}
}

Value *AxeParser::parseDivide(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateUDiv(current, operand);
}

Value *AxeParser::parseModulus(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateURem(current, operand);
}

Value *AxeParser::parseOr(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateOr(current, operand);
}

Value *AxeParser::parseXor(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateXor(current, operand);
}

Value *AxeParser::parseAnd(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateAnd(current, operand);
}

Value *AxeParser::parseNegate(Value *current) {
	Enter("Operand");
		Value *operand = parseOperand(current);
	Exit();
	if (!operand)
		return parseError(ERR_NO_OPERAND);
	return builder->CreateNeg(operand);
}

// Store := Identifier
Value *AxeParser::parseStore(Value *current) {
	Enter("Identifier");
		Value *variable = parseIdentifier(current);
	Exit();
	if (!variable)
		return variable;
	return current;
}

// Load := Identifier
Value *AxeParser::parseLoad() {
	Enter("Identifier");
		Value *variable = parseIdentifier();
	Exit();
	if (!variable)
		return variable;
	return builder->CreateLoad(variable);
}

// Label := 
Function *AxeParser::parseLabel() {
	if (input.currentToken() < 0x41 || input.currentToken() > 0x5b)
		return static_cast<Function *>(parseError(ERR_NO_IDENTIFIER));
	std::string identifier("Lbl_");
	do
		identifier += input.currentToken();
	while (input.nextToken() >= 0x41 && input.currentToken() <= 0x5b);
	PrintF("%s", identifier.c_str());
	if (Function *label = module->getFunction(identifier))
		return label;
	return Function::Create(FunctionType::get(builder->getInt16Ty(),
			builder->getInt16Ty(), false), Function::PrivateLinkage, identifier,
			module);
}

// Identifier := 
Value *AxeParser::parseIdentifier(Value *current) {
	if (input.currentToken() < 0x41 || input.currentToken() > 0x5b)
		return parseConstantIdentifier(current);
	std::string identifier(1, input.currentToken());
	while (input.nextToken() >= 0x41 && input.currentToken() <= 0x5b)
		identifier += input.currentToken();
	PrintF("%s", identifier.c_str());
	Value *variable = variables[identifier];
	if (!variable)
		return parseError(ERR_UNKNOWN_IDENTIFIER);
	if (current)
		builder->CreateStore(current, variable);
	return variable;
}

// ConstantIdentifier := 
Value *AxeParser::parseConstantIdentifier(Value *current, bool optional) {
	const char *prefix;
	if ((input.currentToken() & 0xff00) == 0x6000)
		prefix = "Pic_";
	else if ((input.currentToken() & 0xff00) == 0x6100)
		prefix = "GDB_";
	else if ((input.currentToken() & 0xff00) == 0xaa00)
		prefix = "Str_";
	else
		return parseError(ERR_NO_IDENTIFIER);
	std::string identifier(prefix);
	if ((input.currentToken() & 0xff) >= 0 &&
			(input.currentToken() & 0xff) < 9)
		identifier += (input.currentToken() & 0xff) + '1';
	else if ((input.currentToken() & 0xff) == 9)
		identifier += '0';
	else
		return parseError(ERR_BAD_TOKEN);
	while ((input.nextToken() >= 0x30 && input.currentToken() <= 0x39) ||
			(input.currentToken() >= 0x41 && input.currentToken() <= 0x5b))
		identifier += input.currentToken();
	PrintF("%s", identifier.c_str());
	Value *&variable = variables[identifier];
	if (!current) {
		if (!variable)
			return parseError(ERR_UNKNOWN_IDENTIFIER);
	} else if (!optional || !variable) {
		if (variable)
			return parseError(ERR_SAME_IDENTIFIER);
		Constant *value = dyn_cast<Constant>(current);
		if (!value)
			return parseError(ERR_NO_CONSTANT);
		variable = new GlobalVariable(*module, value->getType(),
				true, GlobalVariable::PrivateLinkage, value, identifier);
	}
	return variable;
}

// DecimalConstant := 
Constant *AxeParser::parseDecimalConstant() {
	if (input.currentToken() < 0x30 || input.currentToken() > 0x39)
		return parseError(ERR_NO_CONSTANT);
	unsigned short value = input.currentToken() - 0x30;
	while (input.nextToken() >= 0x30 && input.currentToken() <= 0x39) {
		unsigned short newValue = value * 10 + input.currentToken() - 0x30;
		if (newValue < value)
			return parseError(ERR_BIG_CONSTANT);
		value = newValue;
	}
	PrintF("%hu", value);
	return builder->getInt16(value);
}

// HexadecimalConstant := [E] HexDigit ( HexDigit ( HexDigit ( HexDigit ) ) )
Constant *AxeParser::parseHexadecimalConstant() {
	if ((input.nextToken() < 0x30 || input.currentToken() > 0x39) &&
			(input.currentToken() < 0x41 || input.currentToken() > 0x46))
		return parseError(ERR_BAD_CONSTANT);
	unsigned short value = 0;
	do {
		if (value & 0xf000)
			return parseError(ERR_BIG_CONSTANT);
		if (input.currentToken() & 0x20)
			value = (value << 4) | (input.currentToken() & 0x0f);
		else
			value = (value << 4) | (input.currentToken() - 0x41 + 10);
	} while ((input.nextToken() >= 0x30 && input.currentToken() <= 0x39) ||
			(input.currentToken() >= 0x41 && input.currentToken() <= 0x46));
	PrintF("%hx", value);
	return builder->getInt16(value);
}

// HexadecimalByte := HexDigit HexDigit
Constant *AxeParser::parseHexadecimalByte() {
	unsigned char value;
	if (input.currentToken() >= 0x30 && input.currentToken() <= 0x39)
		value = (input.currentToken() & 0x0f) << 4;
	else if (input.currentToken() >= 0x41 && input.currentToken() <= 0x46)
		value = (input.currentToken() - 0x41 + 10) << 4;
	else
		return parseError(ERR_BAD_CONSTANT);
	if (input.nextToken() >= 0x30 && input.currentToken() <= 0x39)
		value |= input.currentToken() & 0x0f;
	else if (input.currentToken() >= 0x41 && input.currentToken() <= 0x46)
		value |= input.currentToken() - 0x41 + 10;
	else
		return parseError(ERR_BAD_CONSTANT);
	return builder->getInt8(value);
}

const char *AxeParser::ERR_UNKNOWN = "Unknown instruction.";
const char *AxeParser::ERR_UNSUPPORTED = "Instruction not yet supported.";
const char *AxeParser::ERR_BIG_CONSTANT = "Constant too big.";
const char *AxeParser::ERR_UNKNOWN_IDENTIFIER = "Unknown identifier found.";
const char *AxeParser::ERR_SAME_IDENTIFIER = "Identifier already in use.";
const char *AxeParser::ERR_NO_IDENTIFIER = "Identifier expected.";
const char *AxeParser::ERR_NO_OPERAND = "Operand expected for operator.";
const char *AxeParser::ERR_NO_IF = "If statement expected.";
const char *AxeParser::ERR_NO_LOOP = "Loop statement expected.";
const char *AxeParser::ERR_NO_CONSTANT = "Constant expected.";
const char *AxeParser::ERR_EXTRA_ELSE = "Unexpected else.";
const char *AxeParser::ERR_EXTRA_END = "Unexpected end.";
const char *AxeParser::ERR_NO_END = "Missing an end.";
const char *AxeParser::ERR_WRONG_ARGS = "Wrong number of arguments.";
const char *AxeParser::ERR_BAD_TOKEN = "Invalid constant.";
const char *AxeParser::ERR_BAD_CONSTANT = "Invalid constant.";
const char *AxeParser::ERR_UNKNOWN_CONSTANT = "Unknown constant.";

