#include "InputParser.h"

#include "TokenSet.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/Support/IRBuilder.h"

using namespace llvm;

#include <string>
#include <vector>
#include <map>

class AxeParser {
public:
	AxeParser(InputParser &);
	~AxeParser();

	const char *getName();
	const char *getError();

	bool check();
	bool parse();

private:
	//bool parseNext();
	//bool parseNewLine();
	//bool parseComment();
	//bool parseDecimalConstant();
	//bool parseHexadecimalConstant();
	//bool parseBinaryConstant();
	//bool parseStringConstant();

	bool parseProgram();
	Value *parseBlock(Value *);
	Value *parseStatement(Value *);
	Value *parseIfStatement(Value *);
	Value *parseLoopStatement(Value *);
	Value *parseForStatement(Value *);
	Value *parseLblStatement(Value *);
	Value *parseGotoStatement(Value *);
	Value *parseReturnStatement(Value *);
	Value *parseExpression(Value *);
	Value *parseValue(Value *);
	Value *parseOperation(Value *, bool * = 0);
	Value *parseOperand(Value *, bool * = 0);
	Constant *parseConstantOperation(Value *, bool * = 0);
	Constant *parseConstantOperand(bool * = 0);
	Constant *parseConstantData();
	Value *parseSub(Value *);
	Value *parseOutput(Value *);
	Value *parsePxlOn(Value *);
	Value *parsePxlOff(Value *);
	Value *parsePxlChange(Value *);
	Value *parsePxlTest(Value *);
	Value *parsePtOn(Value *);
	Value *parsePtOff(Value *);
	Value *parsePtChange(Value *);
	Value *parsePtGet(Value *);
	Value *parseGetKey(Value *);
	std::vector<Value *> parseArgs(Value *);
	Value *parseEqual(Value *);
	Value *parseLessThan(Value *);
	Value *parseGreaterThan(Value *);
	Value *parseLessEqual(Value *);
	Value *parseGreaterEqual(Value *);
	Value *parseNotEqual(Value *);
	Value *parseAdd(Value *);
	Value *parseSubtract(Value *);
	Value *parseMultiply(Value *);
	Value *parseDivide(Value *);
	Value *parseModulus(Value *);
	Value *parseOr(Value *);
	Value *parseXor(Value *);
	Value *parseAnd(Value *);
	Value *parseNegate(Value *);
	Value *parseStore(Value *);
	Value *parseLoad();
	Function *parseLabel();
	Value *parseIdentifier(Value * = 0);
	Value *parseConstantIdentifier(Value * = 0, bool = false);
	Constant *parseDecimalConstant();
	Constant *parseHexadecimalConstant();
	Constant *parseHexadecimalByte();

	void clearError();
	Constant *parseError(const char *);

	char name[9];
	const char *error;
	InputParser &input;
	Module *module;
	Function *init, *getkey, *rand, *pause, *normal, *full, *inton, *intoff,
			*diagon, *diagoff, *clrfront, *clrback, *pxlon, *pxloff, *pxlchange,
			*pxltest, *backpxlon, *backpxloff, *backpxlchange, *backpxltest,
			*spriteon, *spriteoff, *spritechange, *spriteget,
			*backspriteon, *backspriteoff, *backspritechange, *backspriteget,
			*dispgraph, *clrhome, *dispstr, *output;
	IRBuilder<> *builder;
	std::map<std::string, Value *> variables;
	unsigned int indent;

	static const char
			*ERR_UNKNOWN,
			*ERR_UNSUPPORTED,
			*ERR_BIG_CONSTANT,
			*ERR_UNKNOWN_IDENTIFIER,
			*ERR_SAME_IDENTIFIER,
			*ERR_NO_IDENTIFIER,
			*ERR_NO_OPERAND,
			*ERR_NO_IF,
			*ERR_NO_LOOP,
			*ERR_NO_CONSTANT,
			*ERR_EXTRA_ELSE,
			*ERR_EXTRA_END,
			*ERR_NO_END,
			*ERR_WRONG_ARGS,
			*ERR_BAD_TOKEN,
			*ERR_BAD_CONSTANT,
			*ERR_UNKNOWN_CONSTANT;
};

inline const char *AxeParser::getError() {
	return error;
}

inline const char *AxeParser::getName() {
	return name;
}

inline void AxeParser::clearError() {
	error = 0;
}

inline Constant *AxeParser::parseError(const char *text) {
	error = text;
	return 0;
}

