#ifndef BUFFER_H
#define BUFFER_H

#include "Sprite.h"

class Buffer : public Sprite {
public:
	enum Mode {
		ON, CHANGE, OFF, GET
	};

	Buffer(int, int, unsigned char *);
	Buffer(int, int, int, int, unsigned char *);
	Buffer(const Buffer &);

	int getX() const;
	int getY() const;

	void move(int, int);
	void moveBy(int, int);

	void draw(Mode, Sprite &);
	void draw(Mode, const Sprite &);
	void draw(Mode, Sprite &) const;

private:
	int x, y;
};

inline Buffer::Buffer(int w, int h, unsigned char *d) : Sprite(w, h, d) { }

inline Buffer::Buffer(int sx, int sy, int w, int h, unsigned char *d) :
		x(sx), y(sy), Sprite(w, h, d) { }

inline Buffer::Buffer(const Buffer &b) : Sprite(b), x(b.x), y(b.y) { }

inline int Buffer::getX() const {
	return x;
}

inline int Buffer::getY() const {
	return y;
}

inline void Buffer::move(int nx, int ny) {
	x = nx;
	y = ny;
}

inline void Buffer::moveBy(int dx, int dy) {
	x += dx;
	y += dy;
}

inline void Buffer::draw(Mode mode, Sprite &sprite) {
	if (mode != GET)
		draw(mode, const_cast<const Sprite &>(sprite));
	else
		const_cast<const Buffer *>(this)->draw(mode, sprite);
}

#endif

