#include "Calculator.h"
#include "Screen.h"

#include "SDL.h"

#include <cstdlib>
#include <cstdio>
using namespace std;

namespace {
	Uint8 ICON[] = {
			0x01, 0x80, 0x1a, 0x58, 0x3e, 0x7c, 0x3f, 0xfc,
			0x3f, 0xfc, 0x3e, 0x7c, 0x1a, 0x58, 0x0a, 0x50,
			0x02, 0x40, 0xfe, 0x7f, 0xcd, 0xa1, 0xb5, 0xaf,
			0x86, 0x63, 0xb5, 0xaf, 0xb5, 0xa1, 0xff, 0xff
	}, ICON_MASK[] = {
			0x01, 0x80, 0x1b, 0xd8, 0x3f, 0xfc, 0x3f, 0xfc,
			0x3f, 0xfc, 0x3f, 0xfc, 0x1b, 0xd8, 0x0b, 0xd0,
			0x03, 0xc0, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
			0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	}, CURSOR[] = {
			0x80, 0x00, 0xc0, 0x00, 0xa0, 0x00, 0x90, 0x00,
			0x88, 0x00, 0x84, 0x00, 0x82, 0x00, 0x81, 0x00,
			0x80, 0x80, 0x80, 0x40, 0x80, 0x20, 0x80, 0x10,
			0x81, 0xf0, 0x89, 0x00, 0x94, 0x80, 0xa4, 0x80,
			0xc2, 0x40, 0x02, 0x40, 0x01, 0x80
	}, CURSOR_MASK[] = {
			0x80, 0x00, 0xc0, 0x00, 0xe0, 0x00, 0xf0, 0x00,
			0xf8, 0x00, 0xfc, 0x00, 0xfe, 0x00, 0xff, 0x00,
			0xff, 0x80, 0xff, 0xc0, 0xff, 0xe0, 0xff, 0xf0,
			0xff, 0xf0, 0xff, 0x00, 0xf7, 0x80, 0xe7, 0x80,
			0xc3, 0xc0, 0x03, 0xc0, 0x01, 0x80
	};
	Uint32 SCREEN_PATTERN_2[] = {
			0x9eab88, 0x97a482,
			0x97a482, 0x909d7d
	}, SCREEN_PATTERN_3[] = {
			0x9eab88, 0x97a482, 0x97a482,
			0x97a482, 0x97a482, 0x97a482,
			0x97a482, 0x97a482, 0x909d7d
	}, SCREEN_PATTERN_4[] = {
			0x9eab88, 0x9eab88, 0x97a482, 0x97a482,
			0x9eab88, 0x9eab88, 0x97a482, 0x97a482,
			0x97a482, 0x97a482, 0x909d7d, 0x909d7d,
			0x97a482, 0x97a482, 0x909d7d, 0x909d7d
	}, SCREEN_PATTERN_5[] = {
			0x9eab88, 0x9eab88, 0x9eab88, 0x97a482, 0x97a482,
			0x9eab88, 0x9eab88, 0x9eab88, 0x97a482, 0x97a482,
			0x9eab88, 0x9eab88, 0x9eab88, 0x97a482, 0x97a482,
			0x97a482, 0x97a482, 0x97a482, 0x909d7d, 0x909d7d,
			0x97a482, 0x97a482, 0x97a482, 0x909d7d, 0x909d7d
	};
	Uint32 *SCREEN_PATTERN[] = {
			SCREEN_PATTERN_2,
			SCREEN_PATTERN_2,
			SCREEN_PATTERN_2,
			SCREEN_PATTERN_3,
			SCREEN_PATTERN_4,
			SCREEN_PATTERN_5
	};
	SDL_Color SCREEN_PALLETE[] = {
			{ 0x97, 0xa4, 0x82, 0x00 },
			{ 0x00, 0x00, 0x00, 0x00 }
	};

//	0x9eab88
//	
//	        0
//	0x9eab88 0x97a482
//	0x97a482 0x909d7d

	typedef struct {
		int refcount;
		bool running, on;
		Uint8 alpha, scale;
		unsigned char *screen;
		SDL_UserEvent *event;
		SDL_mutex *mutex;
		SDL_Thread *thread;
	} Data;

	// This assumes BitsPerPixel <= 8 and palette
	extern "C" SDL_Surface *scale(SDL_Surface *surface, Uint16 amount) {
		SDL_Surface *newSurface = SDL_CreateRGBSurface(SDL_SWSURFACE,
				surface->w * amount, surface->h * amount,
				surface->format->BitsPerPixel,
				surface->format->Rmask,
				surface->format->Gmask,
				surface->format->Bmask,
				surface->format->Amask);
		SDL_SetPalette(newSurface, SDL_LOGPAL, surface->format->palette->colors,
				0, surface->format->palette->ncolors);

		if (SDL_MUSTLOCK(newSurface))
			SDL_LockSurface(newSurface);
		if (SDL_MUSTLOCK(surface))
			SDL_LockSurface(surface);

		Uint8 row, rowOff, col, colOff, bits = surface->format->BitsPerPixel,
				bitsMask = ~(((Uint8)1 << ((Uint8)8 - bits)) - (Uint8)1),
				*rowPtr = (Uint8 *)surface->pixels,
				*destRowPtr = (Uint8 *)newSurface->pixels;
		for (row = 0; row < surface->h; ++row, rowPtr += surface->pitch) {
			for (rowOff = 0; rowOff < amount; ++rowOff,
					destRowPtr += newSurface->pitch) {
				Uint8 bit = 0, *ptr = rowPtr, destBit = 0,
						*destPtr = destRowPtr;
				for (col = 0; col < surface->w; ++col, bit += bits,
						ptr += bit >> 3, bit &= 7) {
					Uint8 current = (*ptr << bit) & bitsMask;
					for (colOff = 0; colOff < amount; ++colOff, destBit += bits,
							destPtr += destBit >> 3, destBit &= 7)
						*destPtr |= current >> destBit;
				}
			}
		}

		if (SDL_MUSTLOCK(surface))
			SDL_UnlockSurface(surface);
		if (SDL_MUSTLOCK(newSurface))
			SDL_UnlockSurface(newSurface);

		return newSurface;
	}

	extern "C" SDL_Surface *createIcon(Uint8 *pixels) {
		SDL_Surface *wrapper = SDL_CreateRGBSurfaceFrom(pixels, 16, 16, 1, 2,
				0, 0, 0, 0);
		SDL_Surface *icon = scale(wrapper, 2);
		SDL_FreeSurface(wrapper);
		return icon;
	}

	extern "C" int thread(void *_) {
		Data *data = (Data *)_;
		SDL_InitSubSystem(SDL_INIT_VIDEO);

		SDL_LockMutex(data->mutex);
			SDL_PushEvent((SDL_Event *)data->event);
		SDL_UnlockMutex(data->mutex);
		Uint8 currentScale = 0;

		SDL_Rect screenRect = {
				SCREEN_PADDING, SCREEN_PADDING, 0, 0
		}, windowRect = {
				0, 0, 0, 0
		};
		SDL_Surface *screen = SDL_CreateRGBSurfaceFrom(data->screen,
				SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_DEPTH, SCREEN_STRIDE,
				0, 0, 0, 0), *window = NULL;
		SDL_SetColors(screen, SCREEN_PALLETE, 0, 2);

		SDL_WM_SetCaption("Axe Emulator", "Axe");
		SDL_Surface *icon = createIcon(ICON), *iconMask = createIcon(ICON_MASK);
		SDL_WM_SetIcon(icon, (Uint8 *)iconMask->pixels);
		SDL_FreeSurface(iconMask);
		SDL_FreeSurface(icon);
		SDL_Cursor *cursor = SDL_CreateCursor(CURSOR, CURSOR_MASK, 16, 19,
				0, 0);
		SDL_SetCursor(cursor);

		SDL_Event event;
		Uint32 next = SDL_GetTicks();
		do {
			SDL_Delay(1);
			while (SDL_PollEvent(&event)) {
				switch (event.type) {
					case SDL_QUIT:
						data->running = false;
						break;
					case SDL_KEYUP:
						break;
					case SDL_KEYDOWN: {
						SDL_LockMutex(data->mutex);
							bool update = true;
							switch (event.key.keysym.sym) {
								case SDLK_PLUS:
								case SDLK_KP_PLUS:
									if (data->alpha != 128)
										data->alpha += 8;
									break;
								case SDLK_MINUS:
								case SDLK_KP_MINUS:
									if (data->alpha)
										data->alpha -= 8;
									else
										update = false;
									break;
								case SDLK_ASTERISK:
								case SDLK_KP_MULTIPLY:
									if (data->scale != 5)
										++data->scale;
									break;
								case SDLK_SLASH:
								case SDLK_KP_DIVIDE:
									if (data->scale != 1)
										--data->scale;
									break;
								case SDLK_F12:
									data->on = !data->on;
									break;
								default:
									update = false;
									break;
							}
						SDL_UnlockMutex(data->mutex);
						if (update)
							SDL_PushEvent((SDL_Event *)data->event);
						break;
					}
					case SDL_USEREVENT: // Update
						SDL_LockMutex(data->mutex);

							data->event->code = false;

							if (data->scale != currentScale) {
								screenRect.w = SCREEN_WIDTH * data->scale;
								screenRect.h = SCREEN_HEIGHT * data->scale;
								windowRect.w = screenRect.w + SCREEN_TWO_PADDING;
								windowRect.h = screenRect.h + SCREEN_TWO_PADDING;

								window = SDL_SetVideoMode(
										windowRect.w, windowRect.h, 32,
										SDL_HWSURFACE | SDL_ANYFORMAT);
								SDL_FillRect(window, NULL, SDL_MapRGB(
										window->format, 0x80, 0x80, 0x80));
								SDL_UpdateRect(window, 0, 0, 0, 0);

								currentScale = data->scale;
							}

							SDL_Surface *scaledScreen = scale(screen,
									currentScale);
							int alpha = data->alpha;
							bool on = data->on;

						SDL_UnlockMutex(data->mutex);

						SDL_SetAlpha(scaledScreen, SDL_SRCALPHA,
								alpha < 64 ? alpha : alpha);
						int patternSize = on ? currentScale : 0;
						Uint32 *patternData = SCREEN_PATTERN[patternSize];
						SDL_Surface *pattern = SDL_CreateRGBSurfaceFrom(
								patternData, patternSize, patternSize,
								32, patternSize << 2,
								0xff0000, 0x00ff00, 0x0000ff, 0);
						SDL_Surface *displayPattern = SDL_DisplayFormat(
								pattern);
						if (displayPattern != NULL) {
							SDL_FreeSurface(pattern);
							pattern = displayPattern;
						}
						SDL_Rect patternRect;
						patternRect.w = patternRect.h = patternSize;
						for (patternRect.y = screenRect.y;
								patternRect.y < screenRect.y + screenRect.h;
								patternRect.y += patternRect.h)
							for (patternRect.x = screenRect.x;
									patternRect.x < screenRect.x + screenRect.w;
									patternRect.x += patternRect.w)
								SDL_BlitSurface(pattern, NULL,
										window, &patternRect);
						SDL_FreeSurface(pattern);

						SDL_BlitSurface(scaledScreen, NULL, window,
								&screenRect);
						SDL_FreeSurface(scaledScreen);
						SDL_UpdateRects(window, 1, &screenRect);
						break;
				}
			}
		} while (data->running);
		data->event->code = true;

		SDL_FreeSurface(screen);
		SDL_FreeCursor(cursor);
		SDL_FreeSurface(window);

		SDL_QuitSubSystem(SDL_INIT_VIDEO);
		return 0;
	}

}

Calculator::Calculator() {
	Data *data = (Data *)malloc(sizeof(Data));
	SDL_Init(0);
	data->refcount = 1;
	data->running = true;
	data->on = true;
	data->alpha = 128;
	data->scale = 2;
	data->screen = (unsigned char *)memset(malloc(SCREEN_SIZE), 0, SCREEN_SIZE);
	data->event = (SDL_UserEvent *)malloc(sizeof(SDL_Event));
	data->event->type = SDL_USEREVENT;
	data->event->code = true;
	data->mutex = SDL_CreateMutex();
	data->thread = SDL_CreateThread(&thread, data);
	_ = data;
}

Calculator::Calculator(const Calculator &other) : _(other._) {
	++((Data *)_)->refcount;
}

Calculator::~Calculator() {
	Data *data = (Data *)_;
	if (!--data->refcount) {
		SDL_WaitThread(data->thread, NULL);
		SDL_DestroyMutex(data->mutex);
		free(data->event);
		free(data->screen);
		free(data);
		SDL_Quit();
	}
}

void Calculator::close() {
	const SDL_QuitEvent quit = { SDL_QUIT };
	SDL_PushEvent((SDL_Event *)&quit);
}

bool Calculator::running() {
	return reinterpret_cast<Data *>(_)->running;
}

Screen::Screen(Calculator &other) :
		Buffer(96, 64, reinterpret_cast<Data *>(other._)->screen), calc(other) {
	SDL_LockMutex(reinterpret_cast<Data *>(calc._)->mutex);
}

Screen::Screen(Calculator &other, int sx, int sy) :
		Buffer(sx, sy, 96, 64, reinterpret_cast<Data *>(other._)->screen),
		calc(other) {
	SDL_LockMutex(reinterpret_cast<Data *>(calc._)->mutex);
}

Screen::~Screen() {
	Data *data = reinterpret_cast<Data *>(calc._);
	if (!data->event->code) {
		data->event->code = true;
		SDL_PushEvent((SDL_Event *)data->event);
	}
	SDL_UnlockMutex(data->mutex);
}

