#include "InputParser.h"

#include <string>

InputParser::InputParser(int argc, char **argv) :
		type(NONE), input(NULL), output(NULL) {
	if (argc > 2)
		output = open(argv[2], "wb", stdout);
	if (argc > 1) {
		input = open(argv[1], "rb", stdin);
		if (input == NULL)
			type = INVALID;
		else if (!checkHeader("**TI83F*\x1A\x0A"))
			type = TEXT;
		else if (validateProgram())
			type = PROGRAM;
		else
			type = INVALID;
	}
}
InputParser::~InputParser() {
	if (input != NULL)
		fclose(input);
	if (output != NULL)
		fclose(output);
}

void InputParser::setOutput(const char *name) {
	if (output == NULL)
		output = open((std::string(name) + ".8xp").c_str(), "wb", stdout);
}

inline FILE *InputParser::open(const char *path, const char *mode, FILE *std) {
	if (path[0] == '-' && path[1] == '\0')
		return std;
	return fopen(path, mode);
}

inline bool InputParser::validateProgram() {
	fseek(input, 42, SEEK_CUR);
	length = 2;
	length = nextWord();
	if (eof()) return false;
	int n = nextWord();
	if ((n != 11 && n != 13) || eof())
		return false;
	if (nextWord() != length - 13 || eof())
		return false;
	if (nextByte() != 5 || eof())
		return false;
	{
		int i;
		for (i = 0; i < 8; i++) {
			name[i] = nextByte();
			if (eof()) return false;
		}
		name[i] = 0;
	}
	if (n == 13 && (nextByte() || eof() ||
			nextByte() & ~0x80 || eof()))
		return false;
	if (nextWord() != length || eof() ||
			nextWord() != length || eof())
		return false;
	return true;
}

inline bool InputParser::checkHeader(const char *header) {
	int c;
	bool valid;
	const char *ptr = header - 1;
	do {
		c = fgetc(input);
		valid = c == *++ptr;
	} while(*ptr && valid);
	if (!valid) {
		if (c != EOF)
			ungetc(c, input);
		while (ptr != header)
			ungetc(*ptr--, input);
	}
	return valid;
}

int InputParser::nextByte() {
	if (!length) {
		cur = EOF;
		return -1;
	}
	--length;
	cur = fgetc(input);
	if (cur == EOF)
		return -1;
	cur &= 0xff;
	return cur;
}

int InputParser::nextWord() {
	if (nextByte() != -1)
		cur = cur | (nextByte() << 8);
	return cur;
}

int InputParser::nextToken() {
	if (nextByte() != -1 && isTwoByteToken(cur))
		cur = (cur << 8) | nextByte();
	return cur;
}

inline bool InputParser::isTwoByteToken(int first) {
	return first == 0xbb || first == 0xaa ||
		(first >= 0x5c && first <= 0x5e) ||
		(first >= 0x60 && first <= 0x63) ||
		first == 0x7e || first == 0xef;
}

