#include <cstdio>

class InputParser {
public:
	enum Type {
		NONE, INVALID, TEXT, PROGRAM
	};

	InputParser(int, char **);
	~InputParser();

	void setOutput(const char *);

	Type getType();
	const char *getName();
	bool hasOutput();

	bool eof();
	int nextByte();
	int nextWord();
	int nextToken();
	int consumeToken();
	bool consumeToken(int);
	int currentToken();
	int eofCurrent();

private:
	static FILE *open(const char *, const char *, FILE *);
	static bool isTwoByteToken(int first);

	bool validateProgram();
	bool checkHeader(const char *);

	Type type;
	char name[9];
	int length, cur;
	FILE *input, *output;
};

inline InputParser::Type InputParser::getType() {
	return type;
}

inline const char *InputParser::getName() {
	return name;
}

inline bool InputParser::hasOutput() {
	return output != NULL;
}

inline bool InputParser::eof() {
	return length == 0 || cur == EOF || feof(input);
}

inline int InputParser::consumeToken() {
	int current = cur;
	nextToken();
	return current;
}

inline bool InputParser::consumeToken(int expected) {
	if (cur != expected)
		return false;
	nextToken();
	return true;
}

inline int InputParser::currentToken() {
	return cur;
}

inline int InputParser::eofCurrent() {
	return cur == EOF;
}

