#include "Sprite.h"
#include "Buffer.h"

unsigned char Sprite::operator[](int index) const {
	if (index < 0 || index >= (width + 7 >> 3) * height)
		return 0;
	return data[index];
}

unsigned char &Sprite::operator[](int index) {
	static unsigned char garbage;
	if (index < 0 || index >= (width + 7 >> 3) * height)
		return garbage = 0;
	return data[index];
}

void Buffer::draw(Mode mode, const Sprite &sprite) {
	switch (mode) {
		case ON:
			for (int yo = 0; yo < sprite.getHeight(); ++yo)
				if (y + yo < getHeight())
					for (int xo = 0; xo < sprite.getWidth(); ++xo)
						if (x + xo < getWidth() && sprite[(xo >> 3) + yo *
								(sprite.getWidth() + 7 >> 3)] & (0x80 >> (xo & 7)))
							(*this)[(x + xo >> 3) + (y + yo) * (getWidth() + 7 >> 3)] |=
									0x80 >> (x + xo & 7);
			break;
		case CHANGE:
			for (int yo = 0; yo < sprite.getHeight(); ++yo)
				if (y + yo < getHeight())
					for (int xo = 0; xo < sprite.getWidth(); ++xo)
						if (x + xo < getWidth() && sprite[(xo >> 3) + yo *
								(sprite.getWidth() + 7 >> 3)] & (0x80 >> (xo & 7)))
							(*this)[(x + xo >> 3) + (y + yo) * (getWidth() + 7 >> 3)] ^=
									0x80 >> (x + xo & 7);
			break;
		case OFF:
			for (int yo = 0; yo < sprite.getHeight(); ++yo) {
				if (y + yo < getHeight()) {
					for (int xo = 0; xo < sprite.getWidth(); ++xo) {
						if (x + xo < getWidth() && sprite[(xo >> 3) + yo *
								(sprite.getWidth() + 7 >> 3)] & (0x80 >> (xo & 7))) {
							(*this)[(x + xo >> 3) + (y + yo) * (getWidth() + 7 >> 3)] |=
									0x80 >> (x + xo & 7);
						} else {
							(*this)[(x + xo >> 3) + (y + yo) * (getWidth() + 7 >> 3)] &=
									~(0x80 >> (x + xo & 7));
						}
					}
				}
			}
			break;
		case GET:
			return;
	}
	x += sprite.getWidth();
}

void Buffer::draw(Mode mode, Sprite &sprite) const {
	if (mode == GET) {
		for (int yo = 0; yo < sprite.getHeight(); ++yo) {
			for (int xo = 0; xo < sprite.getWidth(); ++xo) {
				if ((*this)[(x + xo >> 3) + (y + yo) * (getWidth() + 7 >> 3)] &
						(0x80 >> (x + xo & 7))) {
					sprite[(xo >> 3) + yo * (sprite.getWidth() + 7 >> 3)] |=
							0x80 >> (xo & 7);
				} else {
					sprite[(xo >> 3) + yo * (sprite.getWidth() + 7 >> 3)] &=
							~(0x80 >> (xo & 7));
				}
			}
		}
	}
}

