#include "TokenSet.h"
#include "Font.h"
#include "Buffer.h"

class Tokens : public TokenSet {
public:
	Tokens(const TokenSet **);

	const unsigned char *operator[](int) const;

private:
	const TokenSet **tokens;
};

class Token : public TokenSet {
public:
	Token(const unsigned char *);

	const unsigned char *operator[](int) const;

private:
	const unsigned char *data;
};

namespace {

static const unsigned char ERROR_DATA[] = "?", DMS_DATA[] = "\5DMS",
		DEC_DATA[] = "\5Dec", FRAC_DATA[] = "\5Frac", STORE_DATA[] = "\x1c",
		BOX_PLOT_DATA[] = "Boxplot", LEFT_BRACKET_DATA[] = "\xc1",
		RIGHT_BRACKET_DATA[] = "]", LEFT_BRACE_DATA[] = "{",
		RIGHT_BRACE_DATA[] = "}", RADIANS_DATA[] = "\x15",
		DEGREES_DATA[] = "\x14", INVERSE_DATA[] = "\x11",
		SQUARED_DATA[] = "\x12", TRANSPOSE_DATA[] = "\x16",
		CUBED_DATA[] = "\xd5", LEFT_PAREN_DATA[] = "(",
		RIGHT_PAREN_DATA[] = ")", ROUND_DATA[] = "round(",
		PXL_TEST_DATA[] = "pxl-Test(", AUGMENT_DATA[] = "augment(",
		ROW_SWAP_DATA[] = "rowSwap(", ROW_PLUS_DATA[] = "row+(",
		TIMES_ROW_DATA[] = "*row(", TIMES_ROW_PLUS_DATA[] = "*row+(",
		MAX_DATA[] = "max(", MIN_DATA[] = "min(",
		RECT_RADIUS_DATA[] = "R\5Pr(", RECT_THETA_DATA[] = "R\5P\x5b(", 
		POLAR_X_DATA[] = "P\5Rx(", POLAR_Y_DATA[] = "P\5Ry(",
		MEDIAN_DATA[] = "median(", RANDM_DATA[] = "randM(",
		MEAN_DATA[] = "mean(", SOLVE_DATA[] = "solve(",
		SEQUENCE_DATA[] = "seq(", FNINT_DATA[] = "fnInt(",
		NDERIV_DATA[] = "nDeriv(", FMIN_DATA[] = "fMin(", FMAX_DATA[] = "fMax(",
		SPACE_DATA[] = " ", DOUBLE_QUOTE_DATA[] = "\"", COMMA_DATA[] = ",",
		IMAGINARY_DATA[] = "\xd7", EXCLAMATION_DATA[] = "!",
		CUBIC_REG_DATA[] = "CubicReg ", QUARTIC_REG_DATA[] = "QuartReg ",
		ZERO_DATA[] = "0", ONE_DATA[] = "1", TWO_DATA[] = "2",
		THREE_DATA[] = "3", FOUR_DATA[] = "4", FIVE_DATA[] = "5",
		SIX_DATA[] = "6", SEVEN_DATA[] = "7", EIGHT_DATA[] = "8",
		NINE_DATA[] = "9", PERIOD_DATA[] = ".", LOWERED_E_DATA[] = "\x1b",
		OR_DATA[] = " or ", XOR_DATA[] = " xor ", COLON_DATA[] = ":",
		NEWLINE_DATA[] = "\xd6", AND_DATA[] = " and ",
		A_DATA[] = "A", B_DATA[] = "B", C_DATA[] = "C", D_DATA[] = "D",
		E_DATA[] = "E", F_DATA[] = "F", G_DATA[] = "G", H_DATA[] = "H",
		I_DATA[] = "I", J_DATA[] = "J", K_DATA[] = "K", L_DATA[] = "L",
		M_DATA[] = "M", N_DATA[] = "N", O_DATA[] = "O", P_DATA[] = "P",
		Q_DATA[] = "Q", R_DATA[] = "R", S_DATA[] = "S", T_DATA[] = "T",
		U_DATA[] = "U", V_DATA[] = "V", W_DATA[] = "W", X_DATA[] = "X",
		Y_DATA[] = "Y", Z_DATA[] = "Z", THETA_DATA[] = "\x5b",
		PRGM_DATA[] = "prgm", RADIAN_DATA[] = "Radian",
		DEGREE_DATA[] = "Degree", NORMAL_DATA[] = "Normal", SCI_DATA[] = "Sci",
		ENG_DATA[] = "Eng", FLOAT_DATA[] = "Float",
		EQUAL_DATA[] = "=", LESS_THAN_DATA[] = "<", GREATER_THAN_DATA[] = ">",
		LESS_EQUAL_DATA[] = "\x17", GREATER_EQUAL_DATA[] = "\x19",
		NOT_EQUAL_DATA[] = "\x18", PLUS_DATA[] = "+", MINUS_DATA[] = "-",
		ANS_DATA[] = "Ans", FIX_DATA[] = "Fix ", HORIZ_DATA[] = "Horiz",
		FULL_DATA[] = "Full", FUNC_DATA[] = "Func", PARAM_DATA[] = "Param",
		POLAR_DATA[] = "Polar", SEQ_DATA[] = "Seq",
		IND_AUTO_DATA[] = "IndpntAuto", IND_ASK_DATA[] = "IndpntAsk",
		DEP_AUTO_DATA[] = "DependAuto", DEP_ASK_DATA[] = "DependAsk",
		PLOT_BOX_DATA[] = "\x0a", PLOT_PLUS_DATA[] = "\x0b",
		PLOT_DOT_DATA[] = "\x0c", TIMES_DATA[] = "*", DIVIDE_DATA[] = "/",
		TRACE_DATA[] = "Trace", CLRDRAW_DATA[] = "ClrDraw",
		ZSTANDARD_DATA[] = "ZStandard", ZTRIG_DATA[] = "ZTrig",
		ZBOX_DATA[] = "ZBox", ZOOM_IN_DATA[] = "Zoom In",
		ZOOM_OUT_DATA[] = "Zoom Out", ZSQUARE_DATA[] = "ZSquare",
		ZINTEGER_DATA[] = "ZInteger", ZPREVIOUS_DATA[] = "ZPrevious",
		ZDECIMAL_DATA[] = "ZDecimal", ZOOM_STAT_DATA[] = "ZoomStat",
		ZOOM_RCL_DATA[] = "ZoomRcl", PRINT_SCREEN_DATA[] = "PrintScreen",
		ZOOM_STO_DATA[] = "ZoomSto", TEXT_DATA[] = "Text(",
		NPR_DATA[] = " nPr ", NCR_DATA[] = " nCr ",
		FN_ON_DATA[] = "FnOn ", FN_OFF_DATA[] = "FnOff ",
		STORE_PIC_DATA[] = "StorePic ", RECALL_PIC_DATA[] = "RecallPic ",
		STORE_GDB_DATA[] = "StoreGDB ", RECALL_GDB_DATA[] = "RecallGDB ",
		LINE_DATA[] = "Line(", VERTICAL_DATA[] = "Vertical ",
		PT_ON_DATA[] = "Pt-On(", PT_OFF_DATA[] = "Pt-Off(",
		PT_CHANGE_DATA[] = "Pt-Change(", PXL_ON_DATA[] = "Pxl-On(",
		PXL_OFF_DATA[] = "Pxl-Off(", PXL_CHANGE_DATA[] = "Pxl-Change(",
		SHADE_DATA[] = "Shade(", CIRCLE_DATA[] = "Circle(",
		HORIZONTAL_DATA[] = "Horizontal ", TANGENT_DATA[] = "Tangent(",
		DRAW_INV_DATA[] = "DrawInv ", DRAWF_DATA[] = "DrawF ",
		RAND_DATA[] = "rand", PI_DATA[] = "\xc4", GET_KEY_DATA[] = "getKey",
		SINGLE_QUOTE_DATA[] = "\'", NEGATIVE_DATA[] = "\x1a",
		INT_DATA[] = "int(", ABS_DATA[] = "abs(", DET_DATA[] = "det(",
		IDENTITY_DATA[] = "identity(", DIM_DATA[] = "dim(",
		SUM_DATA[] = "sum(", PROD_DATA[] = "prod(",
		NOT_DATA[] = "not(", IPART_DATA[] = "iPart(", FPART_DATA[] = "fPart(",
		SQUARE_ROOT_DATA[] = "\x10(", CUBE_ROOT_DATA[] = "\x0e\x10(",
		LN_DATA[] = "ln(", EXP_DATA[] = "\xdb^(",
		LOG_DATA[] = "log(", TEN_DATA[] = "\x1d^(",
		SIN_DATA[] = "sin(", ASIN_DATA[] = "sin\x11(",
		COS_DATA[] = "cos(", ACOS_DATA[] = "cos\x11(",
		TAN_DATA[] = "tan(", ATAN_DATA[] = "tan\x11(",
		SINH_DATA[] = "sinh(", ASINH_DATA[] = "sinh\x11(",
		COSH_DATA[] = "cosh(", ACOSH_DATA[] = "cosh\x11(",
		TANH_DATA[] = "tanh(", ATANH_DATA[] = "tanh\x11(",
		IF_DATA[] = "If ", THEN_DATA[] = "Then", ELSE_DATA[] = "Else",
		WHILE_DATA[] = "While ", REPEAT_DATA[] = "Repeat ", FOR_DATA[] = "For(",
		END_DATA[] = "End", RETURN_DATA[] = "Return", LBL_DATA[] = "Lbl ",
		GOTO_DATA[] = "Goto ", PAUSE_DATA[] = "Pause ", STOP_DATA[] = "Stop",
		IS_GREATER_DATA[] = "IS>(", DS_LESS_DATA[] = "DS<(",
		INPUT_DATA[] = "Input ", PROMPT_DATA[] = "Prompt ",
		DISP_DATA[] = "Disp ", DISPGRAPH_DATA[] = "DispGraph",
		OUTPUT_DATA[] = "Output(", CLRHOME_DATA[] = "ClrHome",
		FILL_DATA[] = "Fill(", SORTA_DATA[] = "SortA(", SORTD_DATA[] = "SortD(",
		DISPTABLE_DATA[] = "DispTable", MENU_DATA[] = "Menu(",
		SEND_DATA[] = "Send(", GET_DATA[] = "Get(",
		PLOTS_ON_DATA[] = "PlotsOn ", PLOTS_OFF_DATA[] = "PlotsOff ",
		LIST_DATA[] = "\xdc", PLOT1_DATA[] = "Plot1(", PLOT2_DATA[] = "Plot2(",
		PLOT3_DATA[] = "Plot3(", CARET_DATA[] = "^", X_ROOT_DATA[] = "\xcd\x10",
		ONE_VAR_STATS_DATA[] = "1-Var Stats ",
		TWO_VAR_STATS_DATA[] = "2-Var Stats ",
		LINE_REG_ABX_DATA[] = "LinReg(a+bx) ", EXP_REG_DATA[] = "ExpReg ",
		LN_REG_DATA[] = "LnReg ", PWR_REG_DATA[] = "PwrReg ",
		MED_MED_DATA[] = "Med-Med ", QUAD_REG_DATA[] = "QuadReg ",
		CLRLIST_DATA[] = "ClrList ", CLRTABLE_DATA[] = "ClrTable",
		HISTOGRAM_DATA[] = "Histogram", XY_LINE_DATA[] = "xyLine",
		SCATTER_DATA[] = "Scatter", LINE_REG_AXB_DATA[] = "LinReg(ax+b) ";

static const Token ERROR(ERROR_DATA), DMS(DMS_DATA), DEC(DEC_DATA),
		FRAC(FRAC_DATA), STORE(STORE_DATA), BOX_PLOT(BOX_PLOT_DATA),
		LEFT_BRACKET(LEFT_BRACKET_DATA), RIGHT_BRACKET(RIGHT_BRACKET_DATA),
		LEFT_BRACE(LEFT_BRACE_DATA), RIGHT_BRACE(RIGHT_BRACE_DATA),
		RADIANS(RADIANS_DATA), DEGREES(DEGREES_DATA), INVERSE(INVERSE_DATA),
		SQUARED(SQUARED_DATA), TRANSPOSE(TRANSPOSE_DATA), CUBED(CUBED_DATA),
		LEFT_PAREN(LEFT_PAREN_DATA), RIGHT_PAREN(RIGHT_PAREN_DATA),
		ROUND(ROUND_DATA), PXL_TEST(PXL_TEST_DATA), AUGMENT(AUGMENT_DATA),
		ROW_SWAP(ROW_SWAP_DATA), ROW_PLUS(ROW_PLUS_DATA),
		TIMES_ROW(TIMES_ROW_DATA), TIMES_ROW_PLUS(TIMES_ROW_PLUS_DATA),
		MAX(MAX_DATA), MIN(MIN_DATA),
		RECT_RADIUS(RECT_RADIUS_DATA), RECT_THETA(RECT_THETA_DATA),
		POLAR_X(POLAR_X_DATA), POLAR_Y(POLAR_Y_DATA), MEDIAN(MEDIAN_DATA),
		RANDM(RANDM_DATA), MEAN(MEAN_DATA), SOLVE(SOLVE_DATA),
		SEQUENCE(SEQUENCE_DATA), FNINT(FNINT_DATA), NDERIV(NDERIV_DATA),
		FMIN(FMIN_DATA), FMAX(FMAX_DATA), SPACE(SPACE_DATA),
		DOUBLE_QUOTE(DOUBLE_QUOTE_DATA), COMMA(COMMA_DATA),
		IMAGINARY(IMAGINARY_DATA), EXCLAMATION(EXCLAMATION_DATA),
		CUBIC_REG(CUBIC_REG_DATA), QUARTIC_REG(QUARTIC_REG_DATA),
		ZERO(ZERO_DATA), ONE(ONE_DATA), TWO(TWO_DATA), THREE(THREE_DATA),
		FOUR(FOUR_DATA), FIVE(FIVE_DATA), SIX(SIX_DATA), SEVEN(SEVEN_DATA),
		EIGHT(EIGHT_DATA), NINE(NINE_DATA), PERIOD(PERIOD_DATA),
		LOWERED_E(LOWERED_E_DATA), OR(OR_DATA), XOR(XOR_DATA),
		COLON(COLON_DATA), NEWLINE(NEWLINE_DATA), AND(AND_DATA),
		A(A_DATA), B(B_DATA), C(C_DATA), D(D_DATA), E(E_DATA), F(F_DATA),
		G(G_DATA), H(H_DATA), I(I_DATA), J(J_DATA), K(K_DATA), L(L_DATA),
		M(M_DATA), N(N_DATA), O(O_DATA), P(P_DATA), Q(Q_DATA), R(R_DATA),
		S(S_DATA), T(T_DATA), U(U_DATA), V(V_DATA), W(W_DATA), X(X_DATA),
		Y(Y_DATA), Z(Z_DATA), THETA(THETA_DATA), PRGM(PRGM_DATA),
		RADIAN(RADIAN_DATA), DEGREE(DEGREE_DATA), NORMAL(NORMAL_DATA),
		SCI(SCI_DATA), ENG(ENG_DATA), FLOAT(FLOAT_DATA), EQUAL(EQUAL_DATA),
		LESS_THAN(LESS_THAN_DATA), GREATER_THAN(GREATER_THAN_DATA),
		LESS_EQUAL(LESS_EQUAL_DATA), GREATER_EQUAL(GREATER_EQUAL_DATA),
		NOT_EQUAL(NOT_EQUAL_DATA), PLUS(PLUS_DATA), MINUS(MINUS_DATA),
		ANS(ANS_DATA), FIX(FIX_DATA), HORIZ(HORIZ_DATA), FULL(FULL_DATA),
		FUNC(FUNC_DATA), PARAM(PARAM_DATA), POLAR(POLAR_DATA), SEQ(SEQ_DATA),
		IND_AUTO(IND_AUTO_DATA), IND_ASK(IND_ASK_DATA), DEP_AUTO(DEP_AUTO_DATA),
		DEP_ASK(DEP_ASK_DATA), PLOT_BOX(PLOT_BOX_DATA),
		PLOT_PLUS(PLOT_PLUS_DATA), PLOT_DOT(PLOT_DOT_DATA), TIMES(TIMES_DATA),
		DIVIDE(DIVIDE_DATA), TRACE(TRACE_DATA), CLRDRAW(CLRDRAW_DATA),
		ZSTANDARD(ZSTANDARD_DATA), ZTRIG(ZTRIG_DATA), ZBOX(ZBOX_DATA),
		ZOOM_IN(ZOOM_IN_DATA), ZOOM_OUT(ZOOM_OUT_DATA), ZSQUARE(ZSQUARE_DATA),
		ZINTEGER(ZINTEGER_DATA), ZPREVIOUS(ZPREVIOUS_DATA),
		ZDECIMAL(ZDECIMAL_DATA), ZOOM_STAT(ZOOM_STAT_DATA),
		ZOOM_RCL(ZOOM_RCL_DATA), PRINT_SCREEN(PRINT_SCREEN_DATA),
		ZOOM_STO(ZOOM_STO_DATA), TEXT(TEXT_DATA), NPR(NPR_DATA), NCR(NCR_DATA),
		FN_ON(FN_ON_DATA), FN_OFF(FN_OFF_DATA),
		STORE_PIC(STORE_PIC_DATA), RECALL_PIC(RECALL_PIC_DATA),
		STORE_GDB(STORE_GDB_DATA), RECALL_GDB(RECALL_GDB_DATA),
		LINE(LINE_DATA), VERTICAL(VERTICAL_DATA),
		PT_ON(PT_ON_DATA), PT_OFF(PT_OFF_DATA), PT_CHANGE(PT_CHANGE_DATA),
		PXL_ON(PXL_ON_DATA), PXL_OFF(PXL_OFF_DATA), PXL_CHANGE(PXL_CHANGE_DATA),
		SHADE(SHADE_DATA), CIRCLE(CIRCLE_DATA), HORIZONTAL(HORIZONTAL_DATA),
		TANGENT(TANGENT_DATA), DRAW_INV(DRAW_INV_DATA), DRAWF(DRAWF_DATA),
		RAND(RAND_DATA), PI(PI_DATA), GET_KEY(GET_KEY_DATA),
		SINGLE_QUOTE(SINGLE_QUOTE_DATA), NEGATIVE(NEGATIVE_DATA), INT(INT_DATA),
		ABS(ABS_DATA), DET(DET_DATA), IDENTITY(IDENTITY_DATA), DIM(DIM_DATA),
		SUM(SUM_DATA), PROD(PROD_DATA), NOT(NOT_DATA),
		IPART(IPART_DATA), FPART(FPART_DATA),
		SQUARE_ROOT(SQUARE_ROOT_DATA), CUBE_ROOT(CUBE_ROOT_DATA),
		LN(LN_DATA), EXP(EXP_DATA), LOG(LOG_DATA), TEN(TEN_DATA),
		SIN(SIN_DATA), ASIN(ASIN_DATA), COS(COS_DATA), ACOS(ACOS_DATA),
		TAN(TAN_DATA), ATAN(ATAN_DATA), SINH(SINH_DATA), ASINH(ASINH_DATA),
		COSH(COSH_DATA), ACOSH(ACOSH_DATA), TANH(TANH_DATA), ATANH(ATANH_DATA),
		IF(IF_DATA), THEN(THEN_DATA), ELSE(ELSE_DATA),
		WHILE(WHILE_DATA), REPEAT(REPEAT_DATA), FOR(FOR_DATA), END(END_DATA),
		RETURN(RETURN_DATA), LBL(LBL_DATA), GOTO(GOTO_DATA), PAUSE(PAUSE_DATA),
		STOP(STOP_DATA), IS_GREATER(IS_GREATER_DATA), DS_LESS(DS_LESS_DATA),
		INPUT(INPUT_DATA), PROMPT(PROMPT_DATA), DISP(DISP_DATA),
		DISPGRAPH(DISPGRAPH_DATA), OUTPUT(OUTPUT_DATA), CLRHOME(CLRHOME_DATA),
		FILL(FILL_DATA), SORTA(SORTA_DATA), SORTD(SORTD_DATA),
		DISPTABLE(DISPTABLE_DATA), MENU(MENU_DATA),
		SEND(SEND_DATA), GET(GET_DATA),
		PLOTS_ON(PLOTS_ON_DATA), PLOTS_OFF(PLOTS_OFF_DATA), LIST(LIST_DATA),
		PLOT1(PLOT1_DATA), PLOT2(PLOT2_DATA), PLOT3(PLOT3_DATA),
		CARET(CARET_DATA), X_ROOT(X_ROOT_DATA),
		ONE_VAR_STATS(ONE_VAR_STATS_DATA), TWO_VAR_STATS(TWO_VAR_STATS_DATA),
		LINE_REG_ABX(LINE_REG_ABX_DATA), EXP_REG(EXP_REG_DATA),
		LN_REG(LN_REG_DATA), PWR_REG(PWR_REG_DATA), MED_MED(MED_MED_DATA),
		QUAD_REG(QUAD_REG_DATA), CLRLIST(CLRLIST_DATA), CLRTABLE(CLRTABLE_DATA),
		HISTOGRAM(HISTOGRAM_DATA), XY_LINE(XY_LINE_DATA), SCATTER(SCATTER_DATA),
		LINE_REG_AXB(LINE_REG_AXB_DATA);

static const TokenSet *TIOS_DATA[] = {
	&ERROR, &DMS, &DEC, &FRAC, &STORE, &BOX_PLOT, &LEFT_BRACKET, &RIGHT_BRACKET,
	&LEFT_BRACE, &RIGHT_BRACE, &RADIANS, &DEGREES, &INVERSE, &SQUARED,
	&TRANSPOSE, &CUBED, &LEFT_PAREN, &RIGHT_PAREN, &ROUND, &PXL_TEST,
	&AUGMENT, &ROW_SWAP, &ROW_PLUS, &TIMES_ROW, &TIMES_ROW_PLUS, &MAX, &MIN,
	&RECT_RADIUS, &RECT_THETA, &POLAR_X, &POLAR_Y, &MEDIAN, &RANDM, &MEAN,
	&SOLVE, &SEQUENCE, &FNINT, &NDERIV, &ERROR, &FMIN, &FMAX, &SPACE,
	&DOUBLE_QUOTE, &COMMA, &IMAGINARY, &EXCLAMATION, &CUBIC_REG, &QUARTIC_REG,
	&ZERO, &ONE, &TWO, &THREE, &FOUR, &FIVE, &SIX, &SEVEN, &EIGHT, &NINE,
	&PERIOD, &LOWERED_E, &OR, &XOR, &COLON, &NEWLINE, &AND,
	&A, &B, &C, &D, &E, &F, &G, &H, &I, &J, &K, &L, &M, &N, &O, &P, &Q, &R, &S,
	&T, &U, &V, &W, &X, &Y, &Z, &THETA, &ERROR, &ERROR, &ERROR, &PRGM,
	&ERROR, &ERROR, &ERROR, &ERROR, &RADIAN, &DEGREE, &NORMAL, &SCI, &ENG,
	&FLOAT, &EQUAL, &LESS_THAN, &GREATER_THAN, &LESS_EQUAL, &GREATER_EQUAL,
	&NOT_EQUAL, &PLUS, &MINUS, &ANS, &FIX, &HORIZ, &FULL, &FUNC, &PARAM, &POLAR,
	&SEQ, &IND_AUTO, &IND_ASK, &DEP_AUTO, &DEP_ASK, &ERROR, &PLOT_BOX,
	&PLOT_PLUS, &PLOT_DOT, &TIMES, &DIVIDE, &TRACE, &CLRDRAW, &ZSTANDARD,
	&ZTRIG, &ZBOX, &ZOOM_IN, &ZOOM_OUT, &ZSQUARE, &ZINTEGER, &ZPREVIOUS,
	&ZDECIMAL, &ZOOM_STAT, &ZOOM_RCL, &PRINT_SCREEN, &ZOOM_STO, &TEXT, &NPR,
	&NCR, &FN_ON, &FN_OFF, &STORE_PIC, &RECALL_PIC, &STORE_GDB, &RECALL_GDB,
	&LINE, &VERTICAL, &PT_ON, &PT_OFF, &PT_CHANGE, &PXL_ON, &PXL_OFF,
	&PXL_CHANGE, &SHADE, &CIRCLE, &HORIZONTAL, &TANGENT, &DRAW_INV, &DRAWF,
	&ERROR, &RAND, &PI, &GET_KEY, &SINGLE_QUOTE, &ERROR, &NEGATIVE, &INT, &ABS,
	&DET, &IDENTITY, &DIM, &SUM, &PROD, &NOT, &IPART, &FPART, &ERROR,
	&SQUARE_ROOT, &CUBE_ROOT, &LN, &EXP, &LOG, &TEN, &SIN, &ASIN, &COS, &ACOS,
	&TAN, &ATAN, &SINH, &ASINH, &COSH, &ACOSH, &TANH, &ATANH, &IF, &THEN, &ELSE,
	&WHILE, &REPEAT, &FOR, &END, &RETURN, &LBL, &GOTO, &PAUSE, &STOP,
	&IS_GREATER, &DS_LESS, &INPUT, &PROMPT, &DISP, &DISPGRAPH, &OUTPUT,
	&CLRHOME, &FILL, &SORTA, &SORTD, &DISPTABLE, &MENU, &SEND, &GET,
	&PLOTS_ON, &PLOTS_OFF, &LIST, &PLOT1, &PLOT2, &PLOT3, &ERROR, &CARET,
	&X_ROOT, &ONE_VAR_STATS, &TWO_VAR_STATS, &LINE_REG_ABX, &EXP_REG, &LN_REG,
	&PWR_REG, &MED_MED, &QUAD_REG, &CLRLIST, &CLRTABLE, &HISTOGRAM, &XY_LINE,
	&SCATTER, &LINE_REG_AXB
};

}

const TokenSet &TokenSet::TIOS = Tokens(TIOS_DATA);

Tokens::Tokens(const TokenSet **t) : tokens(t) { }

inline const unsigned char *Tokens::operator[](int token) const {
	return (*tokens[token & 0xff])[token >> 8];
}

Token::Token(const unsigned char *d) : data(d) { }

const unsigned char *Token::operator[](int token) const {
	return data;
}

void TokenSet::render(Buffer::Mode mode, int token, const Font &font,
		Buffer &buffer) const {
	const unsigned char *data = (*this)[token];
	do {
		buffer.draw(mode, font[*data]);
		/*const Sprite &character = font[*data];
		character.renderOr(screen, x, y);
		if ((x += character.getWidth()) >= 96) {
			x = 0;
			y += 8;
		}*/
	} while (*++data);
}

void TokenSet::render(Buffer::Mode mode, int token, const Font &font,
		Buffer &buffer, int x, int y) const {
	buffer.move(x, y);
	render(mode, token, font, buffer);
}

/*void TokenSet::renderOr(int token, const Font &font, Screen &screen, int x,
		int y) const {
	const unsigned char *data = (*this)[token];
	do {
		screen.draw(font[*data]);
		const Sprite &character = font[*data];
		character.renderOr(screen, x, y);
		if ((x += character.getWidth()) >= 96) {
			x = 0;
			y += 8;
		}
	} while (*++data);
}

void TokenSet::renderXor(int token, const Font &font, Screen &screen, int x,
		int y) const {
	const unsigned char *data = (*this)[token];
	do {
		const Sprite &character = font[*data];
		character.renderXor(screen, x, y);
		if ((x += character.getWidth()) >= 96) {
			x = 0;
			y += 8;
		}
	} while (*++data);
}

void TokenSet::renderOff(int token, const Font &font, Screen &screen, int x,
		int y) const {
	const unsigned char *data = (*this)[token];
	do {
		const Sprite &character = font[*data];
		character.renderOff(screen, x, y);
		if ((x += character.getWidth()) >= 96) {
			x = 0;
			y += 8;
		}
	} while (*++data);
}*/

